'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

exports.default = configureDataEntity;

var _uniqid = require('uniqid');

var _uniqid2 = _interopRequireDefault(_uniqid);

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _actions = require('./variables/actions');

var Actions = _interopRequireWildcard(_actions);

var _states = require('./variables/states');

var States = _interopRequireWildcard(_states);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function configureDataEntity() {
  var globalEntityConfig = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  var globalActionConfig = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var defaultEntityConfig = _extends({
    keyExtractor: function keyExtractor(unit) {
      return unit.id;
    },
    keyGenerator: function keyGenerator(unit) {
      return _uniqid2.default.time();
    },
    valueExtractor: function valueExtractor(unit) {
      return unit;
    },
    reducerDefault: [],
    responseValidityTime: null
  }, globalEntityConfig);
  var defaultActionConfig = _extends({
    optimistic: false,
    force: false
  }, globalActionConfig);
  function createDataEntityInstance(reducerName) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var instanceConfig = _extends({
      reducerName: reducerName
    }, defaultEntityConfig, config);
    var requests = [];
    var optimisticKeys = [];

    if (typeof instanceConfig.reducerName !== 'string') {
      throw new Error('Data entity reducer name has to be a string. ' + ('Instead specified ' + instanceConfig.reducerName));
    }
    if (typeof instanceConfig.process !== 'function') {
      throw new Error('Data entity config process property has to be a function');
    }

    function parseAction(type) {
      for (var action in Actions) {
        if (type.startsWith('RDE/' + instanceConfig.reducerName + '/' + action)) {
          return action;
        }
      }
      return null;
    }

    function parseState(type) {
      var action = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : this.parseAction(type);

      if (action !== null) {
        for (var state in States) {
          if (type === 'RDE/' + instanceConfig.reducerName + '/' + action + '_' + state) {
            return state;
          }
        }
      }
      return null;
    }

    function sortByFinishTime(completed) {
      return completed.sort(function (request1, request2) {
        return request1.finishTime < request2.finishTime ? -1 : +(request1.finishTime > request2.finishTime);
      });
    }

    function addUnit(items, unit) {
      return _extends({}, items, _defineProperty({}, instanceConfig.keyExtractor(unit), instanceConfig.valueExtractor(unit)));
    }

    function removeUnit(items, key) {
      return _lodash2.default.omit(items, key);
    }

    function addOptimisticUnit(items, unit) {
      var optimisticKey = instanceConfig.keyExtractor(unit) || instanceConfig.keyGenerator(unit);
      optimisticKeys.push(optimisticKey);
      return _extends({}, items, _defineProperty({}, optimisticKey, unit));
    }

    function newState() {
      var items = {};

      var _requests$reduce = requests.reduce(function (requestObj, request) {
        var key = typeof request.finishTime === 'undefined' ? 'optimistic' : 'succeeded';
        if (key === 'optimistic' && !request.config.optimistic || request.error) {
          return requestObj;
        }
        return _extends({}, requestObj, _defineProperty({}, key, requestObj[key].concat(request)));
      }, { optimistic: [], succeeded: [] }),
          optimistic = _requests$reduce.optimistic,
          succeeded = _requests$reduce.succeeded;

      if (!Array.isArray(instanceConfig.reducerDefault)) {
        items = instanceConfig.reducerDefault;
      } else {
        instanceConfig.reducerDefault.forEach(function (item) {
          addUnit(items, item);
        });
      }
      succeeded = sortByFinishTime(succeeded);
      succeeded.forEach(function (_ref) {
        var action = _ref.action,
            result = _ref.result,
            config = _ref.config;

        switch (action) {
          case Actions.READ_ONE:
          case Actions.CREATE_ONE:
          case Actions.UPDATE_ONE:
            result = [result];
          case Actions.READ_MANY:
          case Actions.CREATE_MANY:
          case Actions.UPDATE_MANY:
            result.forEach(function (unit) {
              items = addUnit(items, unit);
            });
            break;
          case Actions.DELETE_ONE:
          case Actions.DELETE_MANY:
            config.keys.forEach(function (key) {
              items = removeUnit(items, key);
            });
            break;
        }
      });
      optimistic.forEach(function (_ref2) {
        var action = _ref2.action,
            _ref2$config = _ref2.config,
            data = _ref2$config.data,
            keys = _ref2$config.keys;

        switch (action) {
          case Actions.CREATE_ONE:
            data = _lodash2.default.castArray(data);
          case Actions.CREATE_MANY:
            data.forEach(function (unit) {
              items = addOptimisticUnit(items, unit);
            });
            break;
          case Actions.UPDATE_ONE:
            data = _lodash2.default.castArray(data);
          case Actions.UPDATE_MANY:
            keys.forEach(function (key, index) {
              items[key] = _extends({}, items[key], data[index]);
            });
            break;
          case Actions.DELETE_ONE:
          case Actions.DELETE_MANY:
            keys.forEach(function (key) {
              items = removeUnit(items, key);
            });
            break;
        }
      });
      return items;
    }

    function getReducer() {
      return function (state, _ref3) {
        var type = _ref3.type,
            payload = _ref3.payload,
            _ref3$meta = _ref3.meta;
        _ref3$meta = _ref3$meta === undefined ? {} : _ref3$meta;
        var dataEntity = _ref3$meta.dataEntity;

        var action = parseAction(type);
        var requestState = parseState(type, action);
        if (action === Actions.CLEAR) {
          requests = [];
          optimisticKeys = [];
          state = undefined;
        }
        var modified = _lodash2.default.isUndefined(state);
        if (action !== null && requestState !== null) {
          if (requestState === States.START) {
            dataEntity.startTime = +Date.now();
            dataEntity.action = action;
            requests.push(dataEntity);
            modified = dataEntity.config.optimistic;
          } else {
            dataEntity.finishTime = +Date.now();
            dataEntity[requestState === States.FAIL ? 'error' : 'result'] = payload;
            modified = requests.indexOf(dataEntity) !== -1;
          }
          if (!modified) {
            state = _extends({}, state);
          }
        }
        return modified ? newState() : state;
      };
    }

    function getConfig() {
      return _extends({}, instanceConfig);
    }

    function cacheIsValid(action, req) {
      if (instanceConfig.responseValidityTime !== null) {
        var duration = _lodash2.default.isPlainObject(instanceConfig.responseValidityTime) ? instanceConfig.responseValidityTime[action] : instanceConfig.responseValidityTime;
        if (!_lodash2.default.isNumber(duration)) {
          return false;
        }
        return +Date.now() - req.finishTime < duration;
      }
      return false;
    }

    function shouldRequestInner(action, config, request) {
      var valid = false;
      var active = requestHas(action, config, function (req) {
        if (_lodash2.default.isNumber(req.finishTime)) {
          if (_lodash2.default.isUndefined(req.error) && cacheIsValid(action, req)) {
            valid = true;
          }
          request = req;
          return false;
        }
        return true;
      });
      return !active && !valid;
    }

    function isOptimistic(key) {
      return optimisticKeys.indexOf(key) !== -1;
    }

    function shouldRequest(action, config) {
      return shouldRequestInner(action, config);
    }

    function getActionHandler(dispatch) {
      return function (action, config, callback) {
        var meta = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

        var request = null;
        function onFinish(error, result) {
          if (_lodash2.default.isFunction(callback)) {
            callback(error, result);
          }
          if (request !== null) {
            requests.splice(requests.indexOf(request), 1);
          }
        }
        config = _extends({}, defaultActionConfig, config);
        if (config.force || shouldRequestInner(action, config, request)) {
          var dataEntity = { config: config };
          dispatch({ type: getConst(action, States.START), meta: _extends({}, meta, { dataEntity: dataEntity }) });
          var actionRequest = instanceConfig.process(action, dataEntity.config, getConfig());
          if (actionRequest !== null) {
            actionRequest.then(function (result) {
              dispatch({ type: getConst(action, States.SUCCESS), meta: _extends({}, meta, { dataEntity: dataEntity }), payload: result });
              onFinish(null, result);
            }, function (error) {
              dispatch({ type: getConst(action, States.FAIL), meta: _extends({}, meta, { dataEntity: dataEntity }), payload: error });
              onFinish(error);
            });
          } else {
            onFinish(null);
          }
        }
      };
    }

    function getCacheCleaner(dispatch) {
      return function () {
        dispatch({ type: this.getConst(Actions.CLEAR) });
      };
    }

    function getConst(action) {
      var state = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

      return 'RDE/' + instanceConfig.reducerName + '/' + action + (state !== null ? '_' + state : '');
    }

    function requestHas(action, config, matches) {
      return [].concat(_toConsumableArray(requests)).reverse().some(function (request) {
        if (request.action !== action) {
          return false;
        }
        if (!config) {
          return matches(request);
        }
        return _lodash2.default.isEqual(request.config, _extends({}, defaultActionConfig, config)) ? matches(request) : false;
      });
    }

    function isPerforming(action, config) {
      return requestHas(action, config, function (req) {
        return _lodash2.default.isUndefined(req.finishTime);
      });
    }

    function getLastError(action, config) {
      var lastError = null;
      requestHas(action, config, function (req) {
        if (req.error) {
          lastError = req.error;
        }
        return req.error;
      });
      return lastError === null || lastError instanceof Error ? lastError : new Error(lastError);
    }

    return {
      isPerforming: isPerforming,
      isOptimistic: isOptimistic,
      shouldRequest: shouldRequest,
      getCacheCleaner: getCacheCleaner,
      getActionHandler: getActionHandler,
      getReducer: getReducer,
      getConst: getConst,
      getConfig: getConfig,
      getLastError: getLastError,
      parseState: parseState,
      parseAction: parseAction
    };
  }
  return function (entities) {
    var entityKeys = entities;
    if (!Array.isArray(entities)) {
      entityKeys = Object.keys(entities);
    }
    return entityKeys.reduce(function (entitiesObject, key) {
      return _extends({}, entitiesObject, _defineProperty({}, key, createDataEntityInstance(key, entities[key])));
    }, {});
  };
}